#This source code is mainly adapted of PDPMix from R package 'ChiRP'. 
#For more details, see website, https://github.com/stablemarkets/ChiRP developed by Arman Oganisian.

################################################################################
############################ Source Code   #####################################
################################################################################

error_check <- function(func_args, model_type){
  
  x <- all.vars(func_args$formula[[3]])
  y <- all.vars(func_args$formula[[2]]) 
  
  nparams <- length(x) + 1
  
  if(class(func_args$init_k)!='numeric' | length(func_args$init_k)!=1 ){
    stop("ERROR: init_k must be length 1 numeric integer such that 0 < init_k <= n .")
  }else if(func_args$init_k<0){
    stop("ERROR: init_k must be 0 < init_k <= n .")
  }else if( func_args$init_k %% 1 !=0 ){
    stop('ERROR: init_k must be an integer.')
  }else if(  func_args$init_k > nrow(func_args$d_train) ){
    stop('ERROR: init_k cannot be greater than sample size.')
  }
  
  ## Check burnin and iter settings
  if( ! (func_args$burnin %% 1 ==0 & func_args$iter %% 1 ==0) ){
    stop("ERROR: burnin and iter must be integers.")
  }else if( func_args$burnin < 0 | func_args$iter < 0){
    stop("ERROR: burnin and iter must be positive.")
  }else if(func_args$burnin==0){
    warning("Warning (non-fatal): burnin==0. Should specify 0 < burnin < iter.")
  }else if(func_args$iter==0){
    stop('ERROR: iter==0. Should specify 0 < burnin < iter.')
  }else if(! func_args$burnin < func_args$iter){
    stop("ERROR: iter must be greater than burnin.")      
  }
  
  # check training data
  if( !is.data.frame(func_args$d_train) ){
    stop("ERROR: d_train is not a data.frame object.")
  }else if( nrow(func_args$d_train)==0 ){
    stop("ERROR: d_train has no rows.")
  }else if( ncol(func_args$d_train)==0 ){
    stop("ERROR: d_train has no columns.")
  }else if( nrow(func_args$d_train)<=2  ){
    stop('ERROR: d_train should contain at least 3 observations.')
  }else if( nrow(func_args$d_train) < func_args$init_k ){
    stop("ERROR: nrow(d_train) < init_k.")
  }else if( max( ! x %in% colnames(func_args$d_train) )==1 ){
    stop("ERROR: Some covariate specified in formula are not in d_train")
  }else if(  max(apply(func_args$d_train[,x, drop=F], 2, class)!="numeric") ){
    stop("ERROR: All columns of d_train must be numeric.")
  }
  
  for( var in x ){
    unvals <- unique(func_args$d_train[,var])
    nu <- length(unvals)
    if( nu==2 & sum(unvals %in% c(0,1)) != 2 ){
      stop(paste0('ERROR: ',var," in d_train is binary, but not coded as 0-1"))
    }
  }
  
  # check testing data set if supplied.
  if(! is.null(func_args$d_test) ){
    
    if( !is.data.frame(func_args$d_test) ){
      stop("ERROR: d_test is not a data.frame object.")
    }else if( nrow(func_args$d_test)==0 ){
      stop("ERROR: d_test has no rows.")
    }else if( ncol(func_args$d_test)==0 ){
      stop("ERROR: d_test has no columns.")
    }else if( max( ! x %in% colnames(func_args$d_test) )==1 ){
      stop("ERROR: Some covariate specified in formula are not in d_test.")
    }else if(  max(apply(func_args$d_test[,x, drop=F], 2, class)!="numeric") ){
      stop("ERROR: All columns of d_test must be numeric.")
    }
    
  }
  
  if(!is.null(func_args$beta_prior_mean)){
    if( class(func_args$beta_prior_mean) != 'numeric' ){
      stop('ERROR: beta_prior_mean should be a numeric vector.')
    }else if(length(func_args$beta_prior_mean) != (length(x)+1) ){
      stop("ERROR: beta_prior_mean should be length p+1, where p is number of covariates.")
    }
  }
  
  if(!is.null(func_args$beta_prior_var)){
    if( class(func_args$beta_prior_var) != 'numeric' ){
      stop('ERROR: beta_prior_var should be a numeric vector.')
    }else if(length(func_args$beta_prior_var) != (length(x)+1) ){
      stop("ERROR: beta_prior_var should be length p+1, where p is number of covariates.")
    }else if( max(func_args$beta_prior_var<0)==1 ){
      stop("ERROR: beta_prior_var should contain non-zero values since it is a vector of variances.")
    }
  }
  
  if(length(func_args$beta_var_scale)!=1){
    stop("ERROR: beta_var_scale should be a scalar, length 1 numeric.")
  }else if(class(func_args$beta_var_scale)!='numeric'){
    stop("ERROR: beta_var_scale should be numeric.")
  }else if(func_args$beta_var_scale<0){
    stop("ERROR: beta_var_scale should be >0.")
  }
  
  if(!is.numeric(func_args$mu_scale) | length(func_args$mu_scale)!=1 | func_args$mu_scale<=0 ){
    stop("ERROR: mu_scale should be length 1 numeric values greater than 0.")
  }
  
  if(model_type!='fDP'){
    if(!is.numeric(func_args$tau_scale) | length(func_args$tau_scale)!=1 | func_args$tau_scale<=0 ){
      stop("ERROR: tau_scale should be length 1 numeric values greater than 0.")
    }
  }
  
  if(! class(func_args$d_train[,y]) %in% c('numeric', 'integer') ){
    stop("ERROR: outcome is not numeric")
  }
  
  ## model-specific checks
  if(model_type=='ZDP'){
    
    if( length(func_args$phi_y)!=2 ){
      stop("ERROR: phi_y should be a length 2 vector of positive real numbers.")
    }else if( max(func_args$phi_y<0)==1 ){
      stop("ERROR: phi_y contains negative values. Should be positive real numbers")
    }
    
    if( sum(func_args$d_train[,y]==0)==0 ){
      stop("ERROR: outcome does not contain any zeros. ZDPMix() is inappropriate.")
    }
    
    if(!is.null(func_args$gamma_prior_mean)){
      if( class(func_args$gamma_prior_mean) != 'numeric' ){
        stop('ERROR: gamma_prior_mean should be a numeric vector.')
      }else if(length(func_args$gamma_prior_mean) != (length(x)+1) ){
        stop("ERROR: gamma_prior_mean should be length p+1, where p is number of covariates.")
      }
    }
    
    if(!is.null(func_args$gamma_prior_var)){
      
      if( class(func_args$gamma_prior_var) != 'numeric' ){
        stop('ERROR: gamma_prior_var should be a numeric vector.')
      }else if(length(func_args$gamma_prior_var) != (length(x)+1) ){
        stop("ERROR: gamma_prior_var should be length p+1, where p is number of covariates.")
      }else if( max(func_args$gamma_prior_var<0)==1 ){
        stop("ERROR: gamma_prior_var should contain non-zero values since it is a vector of variances.")
      }
      
    }
    
    if( is.null(func_args$prop_sigma_z) ){
      stop('ERROR: Proposal covariance is null.')
    }else if( class(func_args$prop_sigma_z)!='matrix' ){
      stop('ERROR: prop_sigma_z should be a (p+1) X (p+1) numeric covariance matrix')
    }else if(!is.numeric(func_args$prop_sigma_z)){
      stop("ERROR: prop_sigma_z is not numeric.")
    }else if( !isSymmetric(func_args$prop_sigma_z) ){
      stop("ERROR: prop_sigma_z is not symmetric.")
    }else if( det(func_args$prop_sigma_z)==0 ){
      stop("ERROR: prop_sigma_z is not positive definite.")
    }
    
    
  }else if(model_type=='NDP'){
    
    if( length(func_args$phi_y)!=2 ){
      stop("ERROR: phi_y should be a length 2 vector of positive real numbers.")
    }else if( max(func_args$phi_y<0)==1 ){
      stop("ERROR: phi_y contains negative values. Should be positive real numbers")
    }
    
  }else if(model_type=='fDP'){
    
    if( !is.null(func_args$phi_y) & length(func_args$phi_y)!=2 ){
      stop("ERROR: phi_y should be a length 2 vector of positive real numbers.")
    }else if( max(func_args$phi_y<0)==1 ){
      stop("ERROR: phi_y contains negative values. Should be positive real numbers")
    }
    
    if( !is.null(func_args$tau_x) & length(func_args$tau_x)!=2 ){
      stop("ERROR: tau_x should be a length 2 vector of positive real numbers.")
    }else if( max(func_args$phi_y<0)==1 ){
      stop("ERROR: tau_x contains negative values. Should be positive real numbers")
    }
    
  }else if(model_type=='PDP'){
    
    uy <- unique(func_args$d_train[,y])
    if(length(uy)!=2){
      stop("ERROR: outcome is not binary.")
    }else if( sum(uy %in% c(0,1))!=2  ){
      stop("ERROR: outcome must be coded as numeric vector with 1 and 0. 1 indicates events.")
    }
    
    if( is.null(func_args$prop_sigma_b) ){
      stop('ERROR: Proposal covariance is null.')
    }else if( class(func_args$prop_sigma_b)!='matrix' ){
      stop('ERROR: prop_sigma_b should be a (p+1) X (p+1) numeric covariance matrix')
    }else if(!is.numeric(func_args$prop_sigma_b)){
      stop("ERROR: prop_sigma_b is not numeric.")
    }else if( !isSymmetric(func_args$prop_sigma_b) ){
      stop("ERROR: prop_sigma_b is not symmetric.")
    }else if( det(func_args$prop_sigma_b)==0 ){
      stop("ERROR: prop_sigma_b is not positive definite.")
    }
    
  }
  
}


error_check_credible_gradient = function(func_args){
  
  if( class(func_args$x) != 'numeric' ){
    stop("ERROR: x is not numeric")
  }else if( class(func_args$post_draws) != "matrix" ){
    stop("ERROR: post_draws is not a matrix. Should be matrix with dimensions length(x) by #posterior draws ")
  }else if( !is.character(func_args$col_gradient) | !is.character(func_args$col_mean_line) ){
    stop("ERROR: either col_gradient or col_mean_line is not a character vector specifying a particular color")
  }
  
}

# mode clustering 
calc_adjmat <- cluster_assign_mode <- function(c_shell){
  c_shell <- c_shell
  iter<-ncol(c_shell)
  n <- nrow(c_shell)
  
  adjmat<-matrix(0, nrow=n, ncol=n)
  
  ## compute frequency matrix 
  for(i in 1:iter){
    adjmat_i <- matrix(0, nrow=n, ncol=n)
    class <- c_shell[,i]
    
    for(j in 1:n){
      adjmat_i[j,] <- class==class[j]
    }
    adjmat <- adjmat + adjmat_i
  }
  
  ## choose best cluster assignment
  err_vec <- numeric(length = ncol(c_shell))
  for(i in 1:iter){
    adjmat_i <- matrix(0, nrow=n, ncol=n)
    class <- c_shell[,i]
    for(j in 1:n){
      adjmat_i[j,] <- class==class[j]
    } # compute L2 norm.
    err_vec[i] <- sum((adjmat_i - adjmat)^2)
  }
  
  class_mem <- c_shell[, c(1:ncol(c_shell))[ err_vec==min(err_vec) ] ]
  if(is.matrix(class_mem)) class_mem <- class_mem[,1]
  
  adjmat <- adjmat/iter
  return(list(adjmat=adjmat, class_mem=class_mem))
}

# modified mode clustering
calc_adjmat_n <- function(c_shell, nclus=2){
  iter <- ncol(c_shell)
  n <- nrow(c_shell)
  nclus <- nclus
  
  adjmat <- matrix(0, nrow=n, ncol=n)
  
  ## compute frequency matrix 
  for(i in 1:iter){
    adjmat_i <- matrix(0, nrow=n, ncol=n)
    class <- c_shell[,i]
    
    for(j in 1:n){
      adjmat_i[j,] <- class==class[j]
    }
    adjmat <- adjmat + adjmat_i                
  }
  
  adjmat <- adjmat/iter        
  
  ## choose best cluster assignment
  err_vec <- numeric(length = ncol(c_shell))
  for(i in 1:iter){
    adjmat_i <- matrix(0, nrow=n, ncol=n)
    class <- c_shell[,i]
    for(j in 1:n){
      adjmat_i[j,] <- class==class[j]                   
    }
    err_vec[i] <- sum((adjmat_i - adjmat)^2)           
  }
  
  sss <- sort(err_vec, decreasing = FALSE, index.return=TRUE)
  cm <- c_shell[,c(1:ncol(c_shell))[sss$ix]]
  len <- c()
  for(j in 1:iter){ len[j] <- length(table(cm[,j])) }  
  class_mem <- c_shell[,c(1:ncol(c_shell))[len==nclus]]  
  err_mem <- err_vec[len==nclus] 
  return(list(adjmat=adjmat, class_mem=class_mem, err = err_mem, minnc=min(len)))
}


credible_gradient = function(x, post_draws, 
                             col_gradient='lightblue', 
                             col_mean_line='steelblue'){
  
  func_args<-mget(names(formals()),sys.frame(sys.nframe()))
  error_check_credible_gradient(func_args)
  
  colfunc <- colorRampPalette(c("white", col_gradient))
  ci_perc = seq(.99,.01,-.01)
  colvec = colfunc(length(ci_perc))
  names(colvec) = ci_perc
  
  for(i in ci_perc){
    pci = apply(post_draws, 1, quantile, probs=c( (1-i)/2, (1+i)/2  ) )
    polygon(c(x,rev(x)),c(pci[1,],rev(pci[2,])),
            col = colvec[as.character(i)], border = FALSE)
  }
  
  lines(x, apply(post_draws, 1, mean), col=col_mean_line , pch=20 )
}

class_update_train_pdp <- function(n, K , alpha, name_new, uniq_clabs, clabs,
                                   y, x, x_cat_shell, x_num_shell,
                                   cat_idx, num_idx,
                                   beta_shell, beta_new, cat_new, num_new){
  
  ## shells
  c_shell <- matrix(NA, nrow = n, ncol = K + 1)
  colnames(c_shell) <- c(uniq_clabs, name_new)
  
  ## prior for existing cluster
  pr_exist <- matrix(NA, nrow = n, ncol = K)
  colnames(pr_exist) <- uniq_clabs
  
  pr_new <- numeric(length = n)
  
  clabs <- factor(clabs, levels = uniq_clabs)
  for(j in 1:n){
    n_min_j <- table(clabs[-j] )
    pr_exist[j,] <- log( (alpha*(n_min_j==0) + n_min_j)/(j + alpha - 1) )
    pr_new[j] <-  log( (alpha) / (j + alpha - 1) )
  }
  
  
  ## existing clusters
  for(k in uniq_clabs){
    
    lk_exist <- numeric(length = n)
    
    for(p in cat_idx){
      lk_exist <- lk_exist + dbinom(x[ ,p], 1, x_cat_shell[[p]][1,k], T)
    }
    
    for(p in num_idx){
      lk_exist <- lk_exist + dnorm(x[, p],
                                   x_num_shell[[p]][[1]][,k],
                                   sqrt(x_num_shell[[p]][[2]][,k]), T )
    }
    
    eta <- x %*% beta_shell[,k,drop=F]
    eta <- ifelse(eta < -10, -10, ifelse(eta>10, 10, eta))
    
    lk_exist <- lk_exist + dbinom(y, 1,prob = LaplacesDemon::invlogit(eta), T )
    
    
    c_shell[,k] <- lk_exist + pr_exist[,k]
  }
  
  
  ## New clusters
  lk_new <- numeric(length = n)
  
  for(p in cat_idx){
    lk_new <- lk_new + dbinom(x[ ,p], 1, cat_new[p], T)
  }
  
  for(p in num_idx){
    lk_new <- lk_new + dnorm(x[, p], num_new[1,p], sqrt(num_new[2,p]), T)
  }
  
  
  eta <- x %*% t(beta_new)
  eta <- ifelse(eta < -10, -10, ifelse(eta>10, 10, eta))
  
  lk_exist <- lk_exist + dbinom(y, 1,prob = LaplacesDemon::invlogit( eta ), T )
  
  c_shell[,name_new] <- lk_new + pr_new
  
  weights <- t(apply(c_shell, 1, function(x) exp(x)/sum(exp(x))  ))
  
  return(weights)
}

class_update_test_pdp <- function(n, K , alpha, name_new, uniq_clabs, clabs, x,
                                  x_cat_shell, x_num_shell,
                                  cat_idx, num_idx,
                                  cat_new, num_new){
  
  # shells
  c_shell <- matrix(NA, nrow = n, ncol = K + 1)
  colnames(c_shell) <- c(uniq_clabs, name_new)
  
  ## prior for existing cluster
  clabs <- factor(clabs, levels = uniq_clabs)
  pr_exist <- log(table(clabs)/(length(clabs)+alpha))
  
  ## existing clusters
  for(k in uniq_clabs){
    
    lk_exist <- numeric(length = n)
    
    for(p in cat_idx){
      lk_exist <- lk_exist + dbinom(x[ ,p], 1, x_cat_shell[[p]][1,k], T)
    }
    
    for(p in num_idx){
      
      lk_exist <- lk_exist + dnorm(x[, p],
                                   x_num_shell[[p]][[1]][,k],
                                   sqrt(x_num_shell[[p]][[2]][,k]), T )
    }
    
    c_shell[,k] <- lk_exist + pr_exist[k]
  }
  
  
  ## New clusters
  lk_new <- numeric(length = n)
  
  for(p in cat_idx){
    lk_new <- lk_new + dbinom(x[ ,p], 1, cat_new[p], T)
  }
  
  for(p in num_idx){
    lk_new <- lk_new + dnorm(x[, p], num_new[1,p], sqrt(num_new[2,p]), T)
  }
  
  c_shell[,name_new] <- lk_new + log(alpha/(length(clabs)+alpha))
  
  weights <- t(apply(c_shell, 1, function(x) exp(x)/sum(exp(x))  ))
  
  return(weights)
}

post_pred_draw_train_pdp <- function(n, x, pc, beta_shell){
  
  y_pp <- numeric(length = n)
  
  clabs <- unique(pc)
  
  for( k in clabs ){
    y_pp[pc==k] <- LaplacesDemon::invlogit( x[pc==k,, drop=F] %*% beta_shell[,k,drop=F] )
  }
  
  return(y_pp)
}


post_pred_draw_test_pdp <- function(n, x, pc, beta_shell,
                                    name_new, beta_new){
  
  y_pp <- numeric(length = n)
  
  clabs <- unique(pc)
  
  for( k in setdiff(clabs, name_new) ){
    y_pp[pc==k] <- LaplacesDemon::invlogit( x[pc==k,, drop=F] %*% beta_shell[,k,drop=F] )
  }
  
  y_pp[pc==name_new] <- LaplacesDemon::invlogit( x[pc==name_new, , drop=F] %*% t(beta_new) )
  
  return(y_pp)
}

metrop_hastings<-function(x_0, iter=1, log_post_density,
                          proposal_dist = function(x, prop_sigma){ 
                            MASS::mvrnorm(1, mu = x, Sigma = prop_sigma )
                          }, 
                          lower=-Inf, upper=Inf, prop_sigma,
                          ... ){
  for(i in 1:iter){
    # draw from proposal distribution
    x_star <- proposal_dist(x_0, prop_sigma) 
    
    # calculate ratio of conditional posterior densities
    r_num <- do.call(log_post_density, c(list(x_star), list(...)) )
    r_denom <- do.call(log_post_density, c(list(x_0), list(...)) )
    r <- exp(r_num - r_denom)
    rmin<-min(r,1)
    if(is.na(rmin)) browser()
    # accept / reject proposal
    if(rbinom(1,1,rmin)==1){ 
      x_0<-x_star
    }
  }
  
  res<-list(x_0 = x_0, accept_prob = rmin )
  return(res)
}

rcond_post_beta <- function(y, xm, psi, beta_prior_mean, beta_prior_var){
  
  mu_beta <- beta_prior_mean
  v_beta <- diag(beta_prior_var)
  v_beta_inv <- diag(1/beta_prior_var)
  
  xtx <- t(xm)%*%xm
  
  if(length(y)>0 ){
    
    post_cov <- chol2inv(chol( v_beta_inv + (1/psi)*xtx))
    post_mean <- post_cov %*% (v_beta_inv %*% mu_beta + (1/psi)*t(xm)%*%y )
    
  }else{
    post_mean <- mu_beta
    post_cov <- v_beta
  }
  
  draw <- mvtnorm::rmvnorm(n = 1, mean = post_mean, sigma = post_cov)
  return(draw)  
}

cond_post_beta_pdp <- function(beta_vec, y, xm, beta_prior_mean, beta_prior_var){
  
  eta <- xm %*% beta_vec
  eta <- ifelse(eta < -30, -30, ifelse(eta>5, 5, eta))
  
  lk <- sum(dbinom(x = y, size = 1, prob = LaplacesDemon::invlogit(eta), log = T))
  pr <- mvtnorm::dmvnorm(x = beta_vec, beta_prior_mean, diag(beta_prior_var), log = T)
  
  return(lk + pr)
}

rcond_post_psi <- function(beta, y, xm, g1, b1){
  n_k <- length(y)
  if(n_k==0){ 
    shape_k <- g1
    rate_k <- b1
  }else{
    shape_k <- g1 + n_k/2
    rate_k <- .5*(  sum( (y - xm %*% beta)^2 ) )  + b1
  }
  draw <- invgamma::rinvgamma(n = 1, shape = shape_k, rate = rate_k)
  return(draw)
}

rcond_post_mu_x <- function(x_ck, phi_x, lambda, tau ){
  nvec_k <- length(x_ck)
  mu_mean <- (1/(1/tau + nvec_k/phi_x))*(lambda/tau + sum(x_ck)/phi_x)
  mu_sd <- sqrt( (1/tau + nvec_k/phi_x)^(-1) )
  draw <- rnorm(n = 1, mean = mu_mean, sd =  mu_sd)
  return(draw)
}

rcond_post_phi <- function(mu_x, x, g2, b2){
  n_k <- length(x)
  if(n_k==0){ 
    shape_k <- g2
    rate_k <- b2
  }else{
    shape_k <- g2 + n_k/2
    rate_k <- .5*(  sum( (x - mu_x )^2 ) )  + b2
  }
  phi_post <- invgamma::rinvgamma(n = 1, shape = shape_k, rate = rate_k)
  return(phi_post)
}

rcond_post_mu_trt <- function(x_ck){
  draw<-rbeta(n = 1,shape1 = 1 + sum(x_ck), 
              shape2 = 1 + length(x_ck) - sum(x_ck))
  return(draw)
}

cond_post_alpha <- function(alpha, n, K, nvec){
  lik <- lgamma(alpha) - lgamma(n + alpha) + sum(lgamma(nvec + alpha/K)) - K*lgamma(alpha/K)
  pr <- invgamma::dinvgamma(x = alpha, 1, 1,log = T)
  return(lik + pr)
}

cond_post_pz <- function(gamma, y, xp, 
                         gamma_prior_mean,gamma_prior_var){
  
  eta <- xp %*% gamma
  eta <- ifelse(eta < -10, -10, ifelse(eta>10, 10, eta))
  p <- LaplacesDemon::invlogit(eta)
  
  lk <- sum(dbinom(x = y, size = 1, prob = p, log = T))
  pr <- mvtnorm::dmvnorm(x = gamma, gamma_prior_mean, diag(gamma_prior_var), log = T)
  
  return(lk + pr)
}


################################################################################
############################ Code for PDPMix  ##################################
################################################################################

PDPMix.org<-function(d_train, formula, d_test=NULL, burnin=100, iter=1000,
                     beta_prior_mean=NULL, beta_prior_var=NULL,
                     init_k=10, beta_var_scale=1000, mu_scale=1, tau_scale=1,
                     prop_sigma_b = diag(rep(.025, nparams))){
  
  ###------------------------------------------------------------------------###
  #### 0 - Parse User Inputs                                                ####
  ###------------------------------------------------------------------------###
  # error checking user inputs
  if( missing(d_train) ){ stop("ERROR: must specify a training data.frame.") }
  
  x <- all.vars(formula[[3]]) 
  y <- all.vars(formula[[2]]) 
  
  nparams <- length(x) + 1
  
  if(!is.null(d_test)){
    xt <- model.matrix(data=d_test,
                       object= as.formula(paste0('~ ',paste0(x, collapse = '+'))))
    nt <- nrow(xt)
  }
  
  x_type <- vector(length = length(x))
  for(v in 1:length(x)){
    if( length(unique(d_train[, x[v] ]))==2 ){
      x_type[v] <- 'binary'
    }else if(length(unique(d_train[,x[v]]))>2){
      x_type[v] <- 'numeric'
    }
  }
  
  store_l <- iter - burnin
  y <- d_train[,y]
  x_names <- x
  x <- model.matrix(data=d_train, object = formula )
  
  n<-nrow(x)
  
  xall_names <- x_names
  nparams_all <- length(xall_names)
  xall <- d_train[,xall_names]
  
  all_type_map <- c(x_type)
  names(all_type_map) <- c(x_names)
  xall_type <- all_type_map[as.character(xall_names)]
  
  xall_names_bin <- xall_names[xall_type == 'binary']
  xall_names_num <- xall_names[xall_type == 'numeric']
  
  # parse numeric versus binary variables for outcome model
  n_bin_p <- sum(x_type=='binary')
  n_num_p <- sum(x_type=='numeric')
  
  ## calibrate priors
  if(is.null(beta_prior_mean)){
    reg <- glm(data=d_train, formula = formula, family = binomial('probit'))
    beta_prior_mean <- reg$coefficients
  }
  
  if(is.null(beta_prior_var)){
    reg <- glm(data=d_train, formula = formula, family = binomial('probit'))
    beta_prior_var <- beta_var_scale*diag(vcov(reg))
  }
  
  prior_means <- apply(x[,xall_names_num, drop=F], 2, mean)
  names(prior_means) <- xall_names_num
  
  prior_var <- mu_scale*apply(x[,xall_names_num, drop=F], 2, var)
  names(prior_var) <- xall_names_num
  
  g2 <- rep(2, n_num_p)
  names(g2) <- xall_names_num
  
  b2 <- tau_scale*apply(x[,xall_names_num, drop=F], 2, var)
  names(b2) <- xall_names_num
  
  K <- init_k
  a <- 1
  
  ###------------------------------------------------------------------------###
  #### 1 - Create Shells for storing Gibbs Results                          ####
  ###------------------------------------------------------------------------###
  
  alpha <- numeric(length = store_l)
  
  curr_class_list <- 1:K
  class_names <- paste0('c',curr_class_list)
  
  c_shell<-matrix(NA, nrow=n, ncol=1) # shell for indicators
  
  class_shell <- matrix(NA, nrow=n, ncol=store_l)
  
  beta_shell <- matrix(NA, nrow = nparams, ncol = K)
  colnames(beta_shell) <- class_names
  rownames(beta_shell) <- colnames(x)
  
  c_b_shell <- vector(mode = 'list', length = n_bin_p) 
  names(c_b_shell) <- xall_names_bin
  
  c_n_shell <- vector(mode = 'list', length = n_num_p)
  names(c_n_shell) <- xall_names_num
  
  for( p in xall_names_bin ){
    c_b_shell[[p]] <- matrix(NA, nrow = 1, ncol = K)
    
    c_b_shell[[p]][1,] <- rep(.5, K)
    
    colnames(c_b_shell[[p]]) <- class_names
  }
  
  for( p in xall_names_num ){
    c_n_shell[[p]] <- list(matrix(NA, nrow=1, ncol=K),
                           matrix(NA, nrow=1, ncol=K))
    
    c_n_shell[[p]][[2]][1, ] <- rep(0, K)
    c_n_shell[[p]][[2]][1, ] <- rep(20^2, K)
    
    colnames(c_n_shell[[p]][[1]]) <- class_names
    colnames(c_n_shell[[p]][[2]]) <- class_names
  }
  
  c_b_new <- numeric(length = n_bin_p)
  names(c_b_new) <- xall_names_bin
  
  c_n_new <- matrix(NA, nrow = 2, ncol = n_num_p)
  colnames(c_n_new) <- xall_names_num
  
  if(!is.null(d_test)){
    c_b_new_test <- numeric(length = n_bin_p)
    names(c_b_new_test) <- xall_names_bin
    
    c_n_new_test <- matrix(NA, nrow = 2, ncol = n_num_p)
    colnames(c_n_new_test) <- xall_names_num
    
    predictions <- vector(mode = 'list', length = 2)
    predictions[[1]] <- matrix(nrow = n, ncol = store_l)
    predictions[[2]] <- matrix(nrow = nt, ncol = store_l)
    names(predictions) <- c('train','test')
    
    cluster_inds <- vector(mode = 'list', length = 2)
    cluster_inds[[1]] <- matrix(nrow = n, ncol = store_l)
    cluster_inds[[2]] <- matrix(nrow = nt, ncol = store_l)
    names(cluster_inds) <- c('train','test')
  }else{
    predictions <- vector(mode = 'list', length = 1)
    predictions[[1]] <- matrix(nrow = n, ncol = store_l)
    names(predictions) <- c('train')
    
    cluster_inds <- vector(mode = 'list', length = 1)
    cluster_inds[[1]] <- matrix(nrow = n, ncol = store_l)
    names(cluster_inds) <- c('train')
  }
  
  ###------------------------------------------------------------------------###
  #### 2 - Set Initial Values                                               ####
  ###------------------------------------------------------------------------###
  
  # initially assign everyone to one of K clusters randomly with uniform prob
  c_shell[,1] <- sample(x = class_names, size = n,
                        prob = rep(1/K,K), replace = T)
  
  beta_start <- matrix(0, nrow=nparams, ncol=K)
  colnames(beta_start) <- class_names
  
  for(k in class_names){
    beta_shell[,k] <- beta_prior_mean
  }
  
  for(p in xall_names_bin ){
    c_b_shell[[p]][1,] <- rep(1, K)
  }
  
  for(p in xall_names_num ){
    c_n_shell[[p]][[2]][ 1 , ] <- rep(0, K)
    c_n_shell[[p]][[2]][ 1 , ] <- rep(20^2, K)
  }
  
  
  ###------------------------------------------------------------------------###
  #### 3 - Gibbs Sampler                                                    ####
  ###------------------------------------------------------------------------###
  message_seq <- seq(2, iter, round(iter/20) )
  cat(paste0('Iteration ',2,' of ', iter,' (burn-in) \n'))
  message_step <- (iter)/10
  for(i in 2:iter){
    
    ## print message to user
    if( i %% message_step == 1 ) cat(paste0('Iteration ',i,' of ',
                                            iter,
                                            ifelse(i>burnin,' (sampling) ',
                                                   ' (burn-in)') ,
                                            '\n'))
    
    # compute size of each existing cluster
    class_ind<-c_shell[,1]
    nvec<-table(factor(class_ind,levels = class_names) )
    
    ###----------------------------------------------------------------------###
    #### 3.1 - update Concentration Parameter                               ####
    ###----------------------------------------------------------------------###
    
    a_star <- rnorm(n = 1, mean =  a, sd = 1)
    if(a_star>0){
      r_num <- cond_post_alpha(a_star, n, K, nvec)
      r_denom <- cond_post_alpha(a, n, K, nvec)
      r <- exp(r_num - r_denom)
      accept <- rbinom(1,1, min(r,1) )==1
      if(accept){ a <- a_star }
    }
    
    ###----------------------------------------------------------------------###
    #### 3.2 - update parameters conditional on classification              ####
    ###----------------------------------------------------------------------###
    K <- length(unique(class_names))
    for(k in class_names){ 
      
      ck_ind <- class_ind == k
      
      y_ck <- y[ck_ind]
      
      x_ck <- x[ck_ind,, drop=FALSE]
      
      beta_shell[ ,k] <- metrop_hastings(x_0 = beta_shell[ ,k], iter = 1,
                                         log_post_density = cond_post_beta_pdp,
                                         y= y, xm = x_ck,
                                         prop_sigma = prop_sigma_b,
                                         beta_prior_mean = beta_prior_mean,
                                         beta_prior_var = beta_prior_var)[[1]]
      
      for( p in xall_names_bin ){
        c_b_shell[[p]][1, k] <- rcond_post_mu_trt(x_ck = x_ck[, p])
      }
      
      for( p in xall_names_num ){
        c_n_shell[[p]][[1]][1,k] <- rcond_post_mu_x(x_ck = x_ck[, p],
                                                    phi_x = c_n_shell[[p]][[2]][1,k],
                                                    lambda = prior_means[p],
                                                    tau = prior_var[p])
        
        c_n_shell[[p]][[2]][1,k] <- rcond_post_phi(mu_x = c_n_shell[[p]][[1]][1, k],
                                                   x = x_ck[ , p],
                                                   g2 = g2[p], b2 = b2[p])
      }
      
    }
    
    ###----------------------------------------------------------------------###
    #### 3.3 - update classification conditional on parameters              ####
    ###----------------------------------------------------------------------###
    
    beta_new <- mvtnorm::rmvnorm(n = 1, mean = beta_prior_mean,
                                 sigma = diag(beta_prior_var) )
    
    for(p in xall_names_bin){
      c_b_new[p] <- rbeta(n = 1, shape1 = 1, shape2 = 1)
    }
    
    for(p in xall_names_num){
      c_n_new[1, p] <- rnorm(n = 1, prior_means[p], sqrt(prior_var[p]) )
      c_n_new[2, p] <- invgamma::rinvgamma(n = 1, shape = g2[p], rate = b2[p])
    }
    
    name_new <- paste0('c', max(as.numeric(substr(class_names, 2,10))) + 1)
    
    weights <- class_update_train_pdp(n = n, K = length(class_names), alpha = a,
                                      name_new =  name_new,
                                      uniq_clabs = colnames(beta_shell),
                                      clabs = class_ind,
                                      
                                      y = y, x = x,
                                      x_cat_shell = c_b_shell , x_num_shell = c_n_shell,
                                      
                                      cat_idx=xall_names_bin, num_idx=xall_names_num,
                                      
                                      beta_shell = beta_shell,
                                      
                                      beta_new=beta_new,
                                      cat_new = c_b_new, num_new=c_n_new)
    
    c_shell[,1] <- apply(weights, 1, FUN = sample,
                         x=c(colnames(beta_shell), name_new), size=1, replace=T)
    
    ###----------------------------------------------------------------------###
    #### 4.0 - update shell dimensions as clusters die/ are born            ####
    ###----------------------------------------------------------------------###
    
    new_class_names <- unique(c_shell[,1])
    K_new <- length(new_class_names)
    
    new_class_name <- setdiff(new_class_names, class_names)
    if(length(new_class_name)>0){ 
      
      beta_shell <- cbind(beta_shell, t(beta_new) )
      colnames(beta_shell)[K+1] <- new_class_name
      
      for(p in xall_names_bin ){
        c_b_shell[[p]] <- cbind( c_b_shell[[p]], c_b_new[p] )
        colnames(c_b_shell[[p]])[K+1] <- new_class_name
      }
      
      for(p in xall_names_num ){
        c_n_shell[[p]][[1]] <- cbind(c_n_shell[[p]][[1]], c_n_new[1, p])
        colnames(c_n_shell[[p]][[1]])[K+1] <- new_class_name
        
        c_n_shell[[p]][[2]] <- cbind(c_n_shell[[p]][[2]], c_n_new[2, p])
        colnames(c_n_shell[[p]][[2]])[K+1] <- new_class_name
      }
    }
    
    dead_classes <- setdiff(class_names, new_class_names)
    if(length(dead_classes) > 0){
      beta_shell <- beta_shell[,! colnames(beta_shell) %in% dead_classes, drop=F]
      
      for(p in xall_names_bin){
        c_b_shell[[p]] <- c_b_shell[[p]][, ! colnames(c_b_shell[[p]]) %in% dead_classes, drop=F]
      }
      
      for(p in xall_names_num){
        c_n_shell[[p]][[1]] <- c_n_shell[[p]][[1]][, ! colnames(c_n_shell[[p]][[1]]) %in% dead_classes, drop=F]
        c_n_shell[[p]][[2]] <- c_n_shell[[p]][[2]][, ! colnames(c_n_shell[[p]][[2]]) %in% dead_classes, drop=F]
      }
    }
    
    K <- K_new
    class_names <- new_class_names
    
    if( i > burnin ){
      
      ###--------------------------------------------------------------------###
      #### 5.0 - Predictions on a Training and Test set & Store Results     ####
      ###--------------------------------------------------------------------###
      
      if(!is.null(d_test)){
        name_new_test<-paste0('c', max(as.numeric(substr(class_names,2,10)))+1)
        
        for(p in xall_names_bin){
          c_b_new_test[p] <- rbeta(n = 1, shape1 = 1, shape2 = 1)
        }
        
        for(p in xall_names_num){
          c_n_new_test[1, p] <- rnorm(n = 1, prior_means[p], sqrt(prior_var[p]) )
          c_n_new_test[2, p] <- invgamma::rinvgamma(n = 1,shape = g2[p], rate = b2[p])
        }
        
        beta_new <- mvtnorm::rmvnorm(n = 1, mean = beta_prior_mean,
                                     sigma = diag(beta_prior_var) )
        
        weights_test <- class_update_test_pdp(n = nt, K = length(class_names) ,
                                              alpha =a,
                                              name_new = name_new_test,
                                              uniq_clabs = colnames(beta_shell),
                                              clabs = c_shell[,1],
                                              x = xt,
                                              x_cat_shell = c_b_shell,
                                              x_num_shell = c_n_shell,
                                              cat_idx = xall_names_bin,
                                              num_idx = xall_names_num,
                                              cat_new = c_b_new_test,
                                              num_new = c_n_new_test)
        
        c_shell_test <- apply(weights_test, 1, FUN = sample,
                              x=c(colnames(beta_shell), name_new_test),
                              size=1, replace=T)
        
        test_pred <- post_pred_draw_test_pdp(n=nt, x=xt, pc=c_shell_test,
                                             beta_shell = beta_shell,
                                             name_new = name_new_test,
                                             beta_new = beta_new)
        
        train_pred<-post_pred_draw_train_pdp(n=n, x=x, pc=c_shell[,1],
                                             beta_shell = beta_shell)
        
        cluster_inds[['train']][, i-burnin] <- c_shell[,1]
        cluster_inds[['test']][, i-burnin] <- c_shell_test
        
        predictions[['train']][, i-burnin ] <- train_pred
        predictions[['test']][, i-burnin  ] <- test_pred
      }else{
        train_pred<-post_pred_draw_train_pdp(n=n, x=x, pc=c_shell[,1],
                                             beta_shell = beta_shell)
        
        cluster_inds[['train']][, i-burnin] <- c_shell[,1]
        
        predictions[['train']][, i-burnin ] <- train_pred
        
      }
    }
  }
  
  results <- list(cluster_inds=cluster_inds,
                  predictions = predictions)
  cat('Sampling Complete.')
  return(results)
}

################################################################################
############################ Code for CRP  #####################################
################################################################################

PDPMix_CRP <- function (d_train, formula, d_test = NULL, burnin = 100, iter = 1000, 
                        beta_prior_mean = NULL, beta_prior_var = NULL, init_k = 10, 
                        beta_var_scale = 1000, mu_scale = 1, tau_scale = 1, 
                        prop_sigma_b = diag(rep(0.025, nparams))) 
{
  if (missing(d_train)) {
    stop("ERROR: must specify a training data.frame.")
  }
  x <- all.vars(formula[[3]])
  y <- all.vars(formula[[2]])
  nparams <- length(x) + 1
  
  if (!is.null(d_test)) {
    xt <- model.matrix(data = d_test, object = as.formula(paste0("~ ", 
                                                                 paste0(x, collapse = "+"))))
    nt <- nrow(xt)
  }
  x_type <- vector(length = length(x))
  for (v in 1:length(x)) {
    if (length(unique(d_train[, x[v]])) == 2) {
      x_type[v] <- "binary"
    }
    else if (length(unique(d_train[, x[v]])) > 2) {
      x_type[v] <- "numeric"
    }
  }
  store_l <- iter - burnin
  y <- d_train[, y]
  x_names <- x
  x <- model.matrix(data = d_train, object = formula)
  n <- nrow(x)
  xall_names <- x_names
  nparams_all <- length(xall_names)
  xall <- d_train[, xall_names]
  all_type_map <- c(x_type)
  names(all_type_map) <- c(x_names)
  xall_type <- all_type_map[as.character(xall_names)]
  xall_names_bin <- xall_names[xall_type == "binary"]
  xall_names_num <- xall_names[xall_type == "numeric"]
  n_bin_p <- sum(x_type == "binary")
  n_num_p <- sum(x_type == "numeric")
  if (is.null(beta_prior_mean)) {
    reg <- glm(data = d_train, formula = formula, family = binomial("probit"))
    beta_prior_mean <- reg$coefficients
  }
  if (is.null(beta_prior_var)) {
    reg <- glm(data = d_train, formula = formula, family = binomial("probit"))
    beta_prior_var <- beta_var_scale * diag(vcov(reg))
  }
  prior_means <- apply(x[, xall_names_num, drop = F], 2, mean)
  names(prior_means) <- xall_names_num
  prior_var <- mu_scale * apply(x[, xall_names_num, drop = F], 
                                2, var)
  names(prior_var) <- xall_names_num
  g2 <- rep(2, n_num_p)
  names(g2) <- xall_names_num
  b2 <- tau_scale * apply(x[, xall_names_num, drop = F], 2, 
                          var)
  names(b2) <- xall_names_num
  K <- init_k
  a <- 1
  alpha <- numeric(length = store_l)
  curr_class_list <- 1:K
  class_names <- paste0("c", curr_class_list)
  c_shell <- matrix(NA, nrow = n, ncol = 1)
  class_shell <- matrix(NA, nrow = n, ncol = store_l)
  beta_shell <- matrix(NA, nrow = nparams, ncol = K)
  colnames(beta_shell) <- class_names
  rownames(beta_shell) <- colnames(x)
  c_b_shell <- vector(mode = "list", length = n_bin_p)
  names(c_b_shell) <- xall_names_bin
  c_n_shell <- vector(mode = "list", length = n_num_p)
  names(c_n_shell) <- xall_names_num
  for (p in xall_names_bin) {
    c_b_shell[[p]] <- matrix(NA, nrow = 1, ncol = K)
    c_b_shell[[p]][1, ] <- rep(0.5, K)
    colnames(c_b_shell[[p]]) <- class_names
  }
  for (p in xall_names_num) {
    c_n_shell[[p]] <- list(matrix(NA, nrow = 1, ncol = K), 
                           matrix(NA, nrow = 1, ncol = K))
    c_n_shell[[p]][[2]][1, ] <- rep(0, K)
    c_n_shell[[p]][[2]][1, ] <- rep(20^2, K)
    colnames(c_n_shell[[p]][[1]]) <- class_names
    colnames(c_n_shell[[p]][[2]]) <- class_names
  }
  c_b_new <- numeric(length = n_bin_p)
  names(c_b_new) <- xall_names_bin
  c_n_new <- matrix(NA, nrow = 2, ncol = n_num_p)
  colnames(c_n_new) <- xall_names_num
  if (!is.null(d_test)) {
    c_b_new_test <- numeric(length = n_bin_p)
    names(c_b_new_test) <- xall_names_bin
    c_n_new_test <- matrix(NA, nrow = 2, ncol = n_num_p)
    colnames(c_n_new_test) <- xall_names_num
    predictions <- vector(mode = "list", length = 2)
    predictions[[1]] <- matrix(nrow = n, ncol = store_l)
    predictions[[2]] <- matrix(nrow = nt, ncol = store_l)
    names(predictions) <- c("train", "test")
    cluster_inds <- vector(mode = "list", length = 2)
    cluster_inds[[1]] <- matrix(nrow = n, ncol = store_l)
    cluster_inds[[2]] <- matrix(nrow = nt, ncol = store_l)
    names(cluster_inds) <- c("train", "test")
  }
  else {
    predictions <- vector(mode = "list", length = 1)
    predictions[[1]] <- matrix(nrow = n, ncol = store_l)
    names(predictions) <- c("train")
    cluster_inds <- vector(mode = "list", length = 1)
    cluster_inds[[1]] <- matrix(nrow = n, ncol = store_l)
    names(cluster_inds) <- c("train")
  }
  c_shell[, 1] <- sample(x = class_names, size = n, prob = rep(1/K, 
                                                               K), replace = T)
  beta_start <- matrix(0, nrow = nparams, ncol = K)
  colnames(beta_start) <- class_names
  for (k in class_names) {
    beta_shell[, k] <- beta_prior_mean
  }
  for (p in xall_names_bin) {
    c_b_shell[[p]][1, ] <- rep(1, K)
  }
  for (p in xall_names_num) {
    c_n_shell[[p]][[2]][1, ] <- rep(0, K)
    c_n_shell[[p]][[2]][1, ] <- rep(20^2, K)
  }
  message_seq <- seq(2, iter, round(iter/20))
  cat(paste0("Iteration ", 2, " of ", iter, " (burn-in) \n"))
  message_step <- (iter)/10
  for (i in 2:iter) {
    if (i%%message_step == 1) 
      cat(paste0("Iteration ", i, " of ", iter, ifelse(i > 
                                                         burnin, " (sampling) ", " (burn-in)"), "\n"))
    class_ind <- c_shell[, 1]
    nvec <- table(factor(class_ind, levels = class_names))
    a_star <- rnorm(n = 1, mean = a, sd = 1)
    if (a_star > 0) {
      r_num <- cond_post_alpha(a_star, n, K, nvec)
      r_denom <- cond_post_alpha(a, n, K, nvec)
      r <- exp(r_num - r_denom)
      accept <- rbinom(1, 1, min(r, 1)) == 1
      if (accept) {
        a <- a_star
      }
    }
    K <- length(unique(class_names))
    for (k in class_names) {
      ck_ind <- class_ind == k
      y_ck <- y[ck_ind]
      x_ck <- x[ck_ind, , drop = FALSE]
      beta_shell[, k] <- metrop_hastings(x_0 = beta_shell[, 
                                                          k], iter = 1, log_post_density = cond_post_beta_pdp, 
                                         y = y, xm = x_ck, prop_sigma = prop_sigma_b, 
                                         beta_prior_mean = beta_prior_mean, beta_prior_var = beta_prior_var)[[1]]
      for (p in xall_names_bin) {
        c_b_shell[[p]][1, k] <- rcond_post_mu_trt(x_ck = x_ck[, 
                                                              p])
      }
      for (p in xall_names_num) {  
        c_n_shell[[p]][[1]][1, k] <- rcond_post_mu_x(x_ck = x_ck[, 
                                                                 p], phi_x = c_n_shell[[p]][[2]][1, k], lambda = prior_means[p], 
                                                     tau = prior_var[p])
        c_n_shell[[p]][[2]][1, k] <- rcond_post_phi(mu_x = c_n_shell[[p]][[1]][1, 
                                                                               k], x = x_ck[, p], g2 = g2[p], b2 = b2[p])
      }
    }
    beta_new <- mvtnorm::rmvnorm(n = 1, mean = beta_prior_mean, 
                                 sigma = diag(beta_prior_var))
    for (p in xall_names_bin) {
      c_b_new[p] <- rbeta(n = 1, shape1 = 1, shape2 = 1)
    }
    for (p in xall_names_num) {
      c_n_new[1, p] <- rnorm(n = 1, prior_means[p], sqrt(prior_var[p]))
      c_n_new[2, p] <- invgamma::rinvgamma(n = 1, shape = g2[p], 
                                           rate = b2[p])
    }
    name_new <- paste0("c", max(as.numeric(substr(class_names, 
                                                  2, 10))) + 1)
    weights <- class_update_train_pdp_CRP(n = n, K = length(class_names), 
                                          alpha = a, name_new = name_new, uniq_clabs = colnames(beta_shell), 
                                          clabs = class_ind, y = y, x = x, x_cat_shell = c_b_shell, 
                                          x_num_shell = c_n_shell, cat_idx = xall_names_bin, 
                                          num_idx = xall_names_num, beta_shell = beta_shell, 
                                          beta_new = beta_new, cat_new = c_b_new, num_new = c_n_new)
    c_shell[, 1] <- apply(weights, 1, FUN = sample, x = c(colnames(beta_shell), 
                                                          name_new), size = 1, replace = T)
    new_class_names <- unique(c_shell[, 1])
    K_new <- length(new_class_names)
    new_class_name <- setdiff(new_class_names, class_names)
    if (length(new_class_name) > 0) {
      beta_shell <- cbind(beta_shell, t(beta_new))
      colnames(beta_shell)[K + 1] <- new_class_name
      for (p in xall_names_bin) {
        c_b_shell[[p]] <- cbind(c_b_shell[[p]], c_b_new[p])
        colnames(c_b_shell[[p]])[K + 1] <- new_class_name
      }
      for (p in xall_names_num) {
        c_n_shell[[p]][[1]] <- cbind(c_n_shell[[p]][[1]], 
                                     c_n_new[1, p])
        colnames(c_n_shell[[p]][[1]])[K + 1] <- new_class_name
        c_n_shell[[p]][[2]] <- cbind(c_n_shell[[p]][[2]], 
                                     c_n_new[2, p])
        colnames(c_n_shell[[p]][[2]])[K + 1] <- new_class_name
      }
    }
    dead_classes <- setdiff(class_names, new_class_names)
    if (length(dead_classes) > 0) {
      beta_shell <- beta_shell[, !colnames(beta_shell) %in% 
                                 dead_classes, drop = F]
      for (p in xall_names_bin) {
        c_b_shell[[p]] <- c_b_shell[[p]][, !colnames(c_b_shell[[p]]) %in% 
                                           dead_classes, drop = F]
      }
      for (p in xall_names_num) {
        c_n_shell[[p]][[1]] <- c_n_shell[[p]][[1]][, 
                                                   !colnames(c_n_shell[[p]][[1]]) %in% dead_classes, 
                                                   drop = F]
        c_n_shell[[p]][[2]] <- c_n_shell[[p]][[2]][, 
                                                   !colnames(c_n_shell[[p]][[2]]) %in% dead_classes, 
                                                   drop = F]
      }
    }
    K <- K_new
    class_names <- new_class_names
    if (i > burnin) {
      if (!is.null(d_test)) {
        name_new_test <- paste0("c", max(as.numeric(substr(class_names, 
                                                           2, 10))) + 1)
        for (p in xall_names_bin) {
          c_b_new_test[p] <- rbeta(n = 1, shape1 = 1, 
                                   shape2 = 1)
        }
        for (p in xall_names_num) {
          c_n_new_test[1, p] <- rnorm(n = 1, prior_means[p], 
                                      sqrt(prior_var[p]))
          c_n_new_test[2, p] <- invgamma::rinvgamma(n = 1, 
                                                    shape = g2[p], rate = b2[p])
        }
        beta_new <- mvtnorm::rmvnorm(n = 1, mean = beta_prior_mean, 
                                     sigma = diag(beta_prior_var))
        weights_test <- class_update_test_pdp(n = nt, 
                                              K = length(class_names), alpha = a, name_new = name_new_test, 
                                              uniq_clabs = colnames(beta_shell), clabs = c_shell[, 
                                                                                                 1], x = xt, x_cat_shell = c_b_shell, x_num_shell = c_n_shell, 
                                              cat_idx = xall_names_bin, num_idx = xall_names_num, 
                                              cat_new = c_b_new_test, num_new = c_n_new_test)
        c_shell_test <- apply(weights_test, 1, FUN = sample, 
                              x = c(colnames(beta_shell), name_new_test), 
                              size = 1, replace = T)
        test_pred <- post_pred_draw_test_pdp(n = nt, 
                                             x = xt, pc = c_shell_test, beta_shell = beta_shell, 
                                             name_new = name_new_test, beta_new = beta_new)
        train_pred <- post_pred_draw_train_pdp(n = n, 
                                               x = x, pc = c_shell[, 1], beta_shell = beta_shell)
        cluster_inds[["train"]][, i - burnin] <- c_shell[, 
                                                         1]
        cluster_inds[["test"]][, i - burnin] <- c_shell_test
        predictions[["train"]][, i - burnin] <- train_pred
        predictions[["test"]][, i - burnin] <- test_pred
      }
      else {
        train_pred <- post_pred_draw_train_pdp(n = n, 
                                               x = x, pc = c_shell[, 1], beta_shell = beta_shell)
        cluster_inds[["train"]][, i - burnin] <- c_shell[, 
                                                         1]
        predictions[["train"]][, i - burnin] <- train_pred
      }
    }
  }
  results <- list(cluster_inds = cluster_inds, predictions = predictions)
  cat("Sampling Complete.")
  return(results)
}



class_update_train_pdp_CRP <- function(n, K , alpha, name_new, uniq_clabs, clabs,
                                       y, x, x_cat_shell, x_num_shell,
                                       cat_idx, num_idx,
                                       beta_shell, beta_new, cat_new, num_new){
  
  # shells
  c_shell <- matrix(NA, nrow = n, ncol = K + 1)  
  colnames(c_shell) <- c(uniq_clabs, name_new)
  
  ## prior for existing cluster
  pr_exist <- matrix(NA, nrow = n, ncol = K)    
  colnames(pr_exist) <- uniq_clabs
  
  pr_new <- numeric(length = n)
  
  clabs <- factor(clabs, levels = uniq_clabs)   
  
  for(j in 1:n){
    n_min_j <- table(clabs[-j])                 
    pr_exist[j,] <- log( (alpha*(n_min_j==0) + n_min_j)/(j + alpha - 1) )
    pr_new[j] <-  log( (alpha) / (j + alpha - 1) )
  }
  
  for(k in uniq_clabs){
    
    lk_exist <- numeric(length = n)      
    
    for(p in cat_idx){
      lk_exist <- lk_exist + dbinom(x[ ,p], 1, x_cat_shell[[p]][1,k], T)  
    }
  
    for(p in num_idx){ 
      lk_exist <- lk_exist + dnorm(x[, p],         
                                   x_num_shell[[p]][[1]][,k],                                                    
                                   sqrt(x_num_shell[[p]][[2]][,k]), T )    
    }
    
    lk_exist <- lk_exist 
    
    c_shell[,k] <- lk_exist + pr_exist[,k]                                   
  }
  
  ## New clusters
  lk_new <- numeric(length = n)       
  
  for(p in cat_idx){
    lk_new <- lk_new + dbinom(x[ ,p], 1, cat_new[p], T)
  }
  
  for(p in num_idx){
    lk_new <- lk_new + dnorm(x[, p], num_new[1,p], sqrt(num_new[2,p]), T)
  }
  
  lk_exist <- lk_exist 
  
  c_shell[,name_new] <- lk_new + pr_new             
  
  weights <- t(apply(c_shell, 1, function(x) exp(x)/sum(exp(x))  ))  
  
  return(weights)
}


################################################################################
############################ Code for RDMDP ####################################
################################################################################

ddPDPMix_noBP_adj<-function(d_train, formula, d_test=NULL, adj_mat=NULL, burnin=100, iter=1000,
                            beta_prior_mean=NULL, beta_prior_var=NULL, init_a = 1, 
                            init_k=10, beta_var_scale=1000, mu_scale=1, tau_scale=1,
                            prop_sigma_b = diag(rep(.025, nparams))){
  
  
  ###------------------------------------------------------------------------###
  #### 0 - Parse User Inputs                                                ####
  ###------------------------------------------------------------------------###
  if( missing(d_train) ){ stop("ERROR: must specify a training data.frame.") }
  
  x <- all.vars(formula[[3]]) 
  y <- all.vars(formula[[2]]) 
  
  nparams <- length(x) + 1  
  
  if(!is.null(d_test)){ 
    xt <- model.matrix(data=d_test,
                       object= as.formula(paste0('~ ',paste0(x, collapse = '+'))))   
    nt <- nrow(xt)
  }
  
  x_type <- vector(length = length(x))
  for(v in 1:length(x)){                                                            
    if( length(unique(d_train[, x[v] ]))==2 ){
      x_type[v] <- 'binary'
    }else if(length(unique(d_train[,x[v]]))>2){
      x_type[v] <- 'numeric'
    }
  }
  
  store_l <- iter - burnin                                                     
  y <- d_train[,y]
  x_names <- x
  x <- model.matrix(data=d_train, object = formula )                              
  
  n <- nrow(x)
  
  xall_names <- x_names
  nparams_all <- length(xall_names)
  xall <- d_train[,xall_names]                                                      
  
  all_type_map <- c(x_type)
  names(all_type_map) <- c(x_names)
  xall_type <- all_type_map[as.character(xall_names)]       
  
  xall_names_bin <- xall_names[xall_type == 'binary']      
  xall_names_num <- xall_names[xall_type == 'numeric']     
  
  # parse numeric versus binary variables for outcome model
  n_bin_p <- sum(x_type=='binary')                          
  n_num_p <- sum(x_type=='numeric')                         
  
  ## calibrate priors
  if(is.null(beta_prior_mean)){
    reg <- glm(data=d_train, formula = formula, family = binomial('probit'))         
    beta_prior_mean <- reg$coefficients
  }
  
  if(is.null(beta_prior_var)){
    reg <- glm(data=d_train, formula = formula, family = binomial('probit'))         
    beta_prior_var <- beta_var_scale*diag(vcov(reg))                                  
  }
  
  # prior for beta
  prior_means <- apply(x[,xall_names_num, drop=F], 2, mean)             
  names(prior_means) <- xall_names_num
  
  prior_var <- mu_scale*apply(x[,xall_names_num, drop=F], 2, var)        
  names(prior_var) <- xall_names_num
  
  g2 <- rep(2, n_num_p)                                                 
  names(g2) <- xall_names_num
  
  b2 <- tau_scale*apply(x[,xall_names_num, drop=F], 2, var)             
  names(b2) <- xall_names_num
  
  K <- init_k
  a <- init_a  
  
  ###------------------------------------------------------------------------###
  #### 1 - Create Shells for storing Gibbs Results                          ####
  ###------------------------------------------------------------------------###
  
  alpha <- numeric(length = store_l)
  
  curr_class_list <- 1:K
  class_names <- paste0('c',curr_class_list)
  
  c_shell <- matrix(NA, nrow=n, ncol=1) 
  
  class_shell <- matrix(NA, nrow=n, ncol=store_l)                  
  
  beta_shell <- matrix(NA, nrow = nparams, ncol = K)              
  colnames(beta_shell) <- class_names
  rownames(beta_shell) <- colnames(x)
  
  c_b_shell <- vector(mode = 'list', length = n_bin_p)
  names(c_b_shell) <- xall_names_bin
  
  c_n_shell <- vector(mode = 'list', length = n_num_p)           
  names(c_n_shell) <- xall_names_num
  
  for( p in xall_names_bin ){
    c_b_shell[[p]] <- matrix(NA, nrow = 1, ncol = K)
    
    c_b_shell[[p]][1,] <- rep(.5, K)
    
    colnames(c_b_shell[[p]]) <- class_names
  }
  
  for( p in xall_names_num ){
    c_n_shell[[p]] <- list(matrix(NA, nrow=1, ncol=K),
                           matrix(NA, nrow=1, ncol=K))
    
    c_n_shell[[p]][[1]][1, ] <- rep(0, K)                  
    c_n_shell[[p]][[2]][1, ] <- rep(20^2, K)
    
    colnames(c_n_shell[[p]][[1]]) <- class_names
    colnames(c_n_shell[[p]][[2]]) <- class_names
  }
  
  c_b_new <- numeric(length = n_bin_p)
  names(c_b_new) <- xall_names_bin
  
  c_n_new <- matrix(NA, nrow = 2, ncol = n_num_p)
  colnames(c_n_new) <- xall_names_num
  
  #nt=nrow(d_test)
  if(!is.null(d_test)){
    c_b_new_test <- numeric(length = n_bin_p)
    names(c_b_new_test) <- xall_names_bin
    
    c_n_new_test <- matrix(NA, nrow = 2, ncol = n_num_p)
    colnames(c_n_new_test) <- xall_names_num
    
    predictions <- vector(mode = 'list', length = 2)
    predictions[[1]] <- matrix(nrow = n, ncol = store_l)
    predictions[[2]] <- matrix(nrow = nt, ncol = store_l)
    names(predictions) <- c('train','test')
    
    cluster_inds <- vector(mode = 'list', length = 2)
    cluster_inds[[1]] <- matrix(nrow = n, ncol = store_l)
    cluster_inds[[2]] <- matrix(nrow = nt, ncol = store_l)
    names(cluster_inds) <- c('train','test')
  }else{
    predictions <- vector(mode = 'list', length = 1)
    predictions[[1]] <- matrix(nrow = n, ncol = store_l)
    names(predictions) <- c('train')
    
    cluster_inds <- vector(mode = 'list', length = 1)
    cluster_inds[[1]] <- matrix(nrow = n, ncol = store_l)
    names(cluster_inds) <- c('train')
  }
  
  LL <- matrix(NA, nrow=1, ncol= iter)
  LLmat <- matrix(NA, nrow=n, ncol=iter)
  BIC <- matrix(NA, nrow=1, ncol= iter)
  AIC <- matrix(NA, nrow=1, ncol= iter)
  
  Beta <- vector(mode='list', length = length(xall_names)  ) 
  for(j in 1:length(xall_names)){
    Beta[[j]] <- matrix(NA, nrow=500, ncol=iter)   
  }
  
  ###------------------------------------------------------------------------###
  #### 2 - Set Initial Values                                               ####
  ###------------------------------------------------------------------------###
  
  # initially assign everyone to one of K clusters randomly with uniform prob
  c_shell[,1] <- sample(x = class_names, size = n,
                        prob = rep(1/K,K), replace = T)               
  
  beta_start <- matrix(0, nrow=nparams, ncol=K)
  colnames(beta_start) <- class_names
  
  for(k in class_names){
    beta_shell[,k] <- beta_prior_mean                               
  }
  
  for(p in xall_names_bin ){                                        
    c_b_shell[[p]][1,] <- rep(1, K)
  }
  
  for(p in xall_names_num ){                                      
    c_n_shell[[p]][[1]][ 1 , ] <- rep(0, K)                         
    c_n_shell[[p]][[2]][ 1 , ] <- rep(20^2, K)
  }
  
  
  ###------------------------------------------------------------------------###
  #### 3 - Gibbs Sampler                                                    ####
  ###------------------------------------------------------------------------###
  message_seq <- seq(2, iter, round(iter/20) )  
  cat(paste0('Iteration ',2,' of ', iter,' (burn-in) \n'))
  message_step <- (iter)/10
  
  for(i in 2:iter){  
    
    ## print message to user
    if( i %% message_step == 1 ) cat(paste0('Iteration ',i,' of ',
                                            iter,
                                            ifelse(i>burnin,' (sampling) ',
                                                   ' (burn-in)') ,
                                            '\n'))
    
    # compute size of each existing cluster
    class_ind <- c_shell[,1]                                  
    nvec <- table(factor(class_ind,levels = class_names) )    
    table(class_ind)
    ###----------------------------------------------------------------------###
    #### 3.1 - update Concentration Parameter                               ####
    ###----------------------------------------------------------------------###
    a_star <- rnorm(n = 1, mean =  a, sd = 1)          
    if(a_star>0){
      r_num <- cond_post_alpha(a_star, n, K, nvec)                                     
      r_denom <- cond_post_alpha(a, n, K, nvec)
      r <- exp(r_num - r_denom)
      accept <- rbinom(1,1, min(r,1) )==1
      if(accept){ a <- a_star }
    }
    
    ###----------------------------------------------------------------------###
    #### 3.2 - update parameters conditional on classification              ####
    ###----------------------------------------------------------------------###
    K <- length(unique(class_names))
    D <- dim(x)[2]
 
    for(k in class_names){ 
      
      ck_ind <- class_ind == k
      
      y_ck <- y[ck_ind]                                                      
      x_ck <- x[ck_ind,]  
      
      if(is.matrix(x_ck)){
        nn  <- nrow(x_ck)
      } else {
        nn <- 1
      }
      
      # Conjugate prior on the coefficients: beta_shell ~ N (beta_shell_0, Q_0)
      beta_shell_0 <- beta_prior_mean
      
      Q_0 <- diag(1000, D)
      prec_0 <- solve(Q_0)
      
      if(is.matrix(x_ck)){
        V <- solve(prec_0 + crossprod(x_ck, x_ck))
      } else {
        V <- solve(prec_0 + x_ck*x_ck)
      }
      
      
      z <- rep(0, nn)
      N1 <- sum(y_ck)
      N0 <- nn - N1
      
      require(mvtnorm)
      require(truncnorm)
      
      beta_shell_iter = beta_shell[,k]
      
      mu_z <- x_ck %*% beta_shell_iter
      
      # Draw latent variable z from its full conditional: z | \theta, y, X
      if(sum(y_ck==0)>0){
        z[y_ck==0] <- rtruncnorm(N0, mean = mu_z[y_ck==0], sd = 1, a = -Inf, b = 0)
      }
      if(sum(y_ck==1)>0){
        z[y_ck==1] <- rtruncnorm(N1, mean = mu_z[y_ck==1], sd = 1, a = 0, b = Inf)
      }
      
      if(is.matrix(x_ck)){
        M <- V %*% (prec_0 %*% beta_shell_0 + crossprod(x_ck, z))
        # Draw variable \theta from its full conditionals: \theta | x, X
        beta_shell_iter <- c(rmvnorm(1, M, V))
        
      } else {
        M <- V %*% (prec_0 %*% beta_shell_0 + x_ck*z)
        # Draw variable \theta from its full conditionals: \theta | x, X
        beta_shell_iter <- c(rnorm(1, M, V))
      }
      
      beta_shell[,k] <- beta_shell_iter 
      
      for( p in xall_names_bin ){    
        c_b_shell[[p]][1, k] <- rcond_post_mu_trt(x_ck = x_ck[, p])
      }

      for( p in xall_names_num ){       
        
        if(is.matrix(x_ck)){
          c_n_shell[[p]][[1]][1,k] <- rcond_post_mu_x(x_ck = x_ck[,p],       
                                                      phi_x = c_n_shell[[p]][[2]][1,k],
                                                      lambda = prior_means[p],  
                                                      tau = prior_var[p])      
          
          c_n_shell[[p]][[2]][1,k] <- rcond_post_phi(mu_x = c_n_shell[[p]][[1]][1, k],     
                                                     x = x_ck[,p],         
                                                     g2 = g2[p], b2 = b2[p]) 
        } else {
          c_n_shell[[p]][[1]][1,k] <- rcond_post_mu_x(x_ck = x_ck[xall_names_num],       
                                                      phi_x = c_n_shell[[p]][[2]][1,k],
                                                      lambda = prior_means[p],  
                                                      tau = prior_var[p])       
          
          c_n_shell[[p]][[2]][1,k] <- rcond_post_phi(mu_x = c_n_shell[[p]][[1]][1, k],     
                                                     x = x_ck[xall_names_num],         
                                                     g2 = g2[p], b2 = b2[p]) 
        }
        
      }
    }
    
    
    ###----------------------------------------------------------------------###
    #### 3.3 - update classification conditional on parameters              ####
    ###----------------------------------------------------------------------###
    
    
    beta_new <- mvtnorm::rmvnorm(n = 1, mean = beta_prior_mean,
                                 sigma = diag(beta_prior_var) )             
    
    for(p in xall_names_bin){ 
      c_b_new[p] <- rbeta(n = 1, shape1 = 1, shape2 = 1)                     
    }
    
    for(p in xall_names_num){
      c_n_new[1, p] <- rnorm(n = 1, prior_means[p], sqrt(prior_var[p]) )       
      c_n_new[2, p] <- invgamma::rinvgamma(n = 1, shape = g2[p], rate = b2[p])
    }
    
    name_new <- paste0('c', max(as.numeric(substr(class_names, 2,10))) + 1)    
    
    K = length(class_names); alpha = a; 
    uniq_clabs = colnames(beta_shell); clabs = class_ind;
    x_cat_shell = c_b_shell; x_num_shell = c_n_shell;
    cat_idx=xall_names_bin; num_idx=xall_names_num;
    cat_new = c_b_new; num_new=c_n_new
    
    c_shell <- matrix(NA, nrow = n, ncol = K + 1)
    colnames(c_shell) <- c(uniq_clabs, name_new)
    
    ## prior for existing cluster
    pr_exist <- matrix(NA, nrow = n, ncol = K)    
    colnames(pr_exist) <- uniq_clabs
    pr_new <- numeric(length = n)
    clabs <- factor(clabs, levels = uniq_clabs)   
    
    
    for(j in 1:n){
      n_min_j <- table(clabs[-j])
      pr_exist[j,] <- log( (alpha*(n_min_j==0) + n_min_j)/(j + alpha - 1) )
      pr_new[j] <-  log( (alpha) / (j + alpha - 1) )
    }
    
    pr_exist <- pr_exist - min(pr_exist)
    pr_new <- pr_new - min(pr_new)  
    
    for(k in uniq_clabs){
      lk_exist <- numeric(length = n)    
      for(p in cat_idx){ 
        lk_exist <- lk_exist + dbinom(x[ ,p], 1, x_cat_shell[[p]][1,k], T)  
      }
      for(p in num_idx){ 
        lk_exist <- lk_exist + dnorm(x[, p],         
                                     x_num_shell[[p]][[1]][,k],
                                     sqrt(x_num_shell[[p]][[2]][,k]), T )   
      }
      c_shell[,k] <- lk_exist + pr_exist[,k]           
    }
    
    ## New cluster
    lk_new <- numeric(length = n)     
    for(p in cat_idx){
      lk_new <- lk_new + dbinom(x[ ,p], 1, cat_new[p], T)
    }
    for(p in num_idx){
      lk_new <- lk_new + dnorm(x[, p], num_new[1,p], sqrt(num_new[2,p]), T)
    }
    c_shell[,name_new] <- lk_new + pr_new 
    
    weights <- t(apply(c_shell, 1, function(x) exp(x)/sum(exp(x))  ))  
    ww <- weights 
    
    for(jj in 1:n){
      rww <- ww[jj,]
      aa <- adj_mat[jj,]             
      neighbor_c <- clabs[aa==1]  
      tf <- c(colnames(beta_shell), name_new) %in% sort(unique(neighbor_c)) 
      rww[which(tf[-length(tf)]==FALSE)] <- 0      
      rww <- rww/sum(rww)                          
      rww.new <- rww[ncol(ww)]
      rww[ncol(ww)] <- ww[jj,ncol(ww)]                
      rww[which(tf[-length(tf)]==TRUE)] <- rww[which(tf[-length(tf)]==TRUE)] + (rww.new-ww[jj,ncol(ww)])/length(which(tf==TRUE))  
      ww[jj, ] <- rww
    }

    c_shell[,1] <- apply(ww, 1, FUN = sample,                                      
                         x=c(colnames(beta_shell), name_new), size=1, replace=T)   

    ###----------------------------------------------------------------------###
    #### 4.0 - update shell dimensions as clusters die/ are born            ####
    ###----------------------------------------------------------------------###
    
    new_class_names <- unique(c_shell[,1])                         
    K_new <- length(new_class_names)
    
    # Grow shells by adding labels/parameters of new clusters
    new_class_name <- setdiff(new_class_names, class_names)
    if(length(new_class_name)>0){
      
      beta_shell <- cbind(beta_shell, t(beta_new) )
      colnames(beta_shell)[K+1] <- new_class_name
      
      for(p in xall_names_bin ){
        c_b_shell[[p]] <- cbind( c_b_shell[[p]], c_b_new[p] )
        colnames(c_b_shell[[p]])[K+1] <- new_class_name
      }
      
      for(p in xall_names_num ){
        c_n_shell[[p]][[1]] <- cbind(c_n_shell[[p]][[1]], c_n_new[1, p])
        colnames(c_n_shell[[p]][[1]])[K+1] <- new_class_name
        
        c_n_shell[[p]][[2]] <- cbind(c_n_shell[[p]][[2]], c_n_new[2, p])
        colnames(c_n_shell[[p]][[2]])[K+1] <- new_class_name
      }
    }
    
    # shrink shells by deleting labels of clusters that have died.
    dead_classes <- setdiff(class_names, new_class_names)                       
    if(length(dead_classes) > 0){ 
      beta_shell <- beta_shell[,! colnames(beta_shell) %in% dead_classes, drop=F]
      
      for(p in xall_names_bin){
        c_b_shell[[p]] <- c_b_shell[[p]][, ! colnames(c_b_shell[[p]]) %in% dead_classes, drop=F]
      }
      
      for(p in xall_names_num){
        c_n_shell[[p]][[1]] <- c_n_shell[[p]][[1]][, ! colnames(c_n_shell[[p]][[1]]) %in% dead_classes, drop=F]
        c_n_shell[[p]][[2]] <- c_n_shell[[p]][[2]][, ! colnames(c_n_shell[[p]][[2]]) %in% dead_classes, drop=F]
      }
    }
    
    K <- K_new
    class_names <- new_class_names
    
    ###----------------------------------------------------------------------###
    #### 4.1 - save entropy/loglikelihood based on the updated classees      ####
    ###----------------------------------------------------------------------###    
    
    library(stringr)
    
    loglik <- matrix(NA, nrow=length(y))
    
    for(k in class_names){ 
      ck_ind <- c_shell[,1] == k          
      y_ck <- y[ck_ind]                   
      x_ck <- x[ck_ind,] 
      n_k <- length(y_ck)
      
      lk_exist <- numeric(length = n_k)      
      
      for(p in xall_names_num){ 
        if(n_k > 1){
          lk_exist <- lk_exist + dnorm(x_ck[,p],                 
                                       c_n_shell[[p]][[1]][,k],
                                       sqrt(c_n_shell[[p]][[2]][,k]), T )   
        } else {
          lk_exist <- lk_exist + dnorm(x_ck[-1],         
                                       c_n_shell[[p]][[1]][,k],
                                       sqrt(c_n_shell[[p]][[2]][,k]), T )   
        }
      }
      
      if(which(class_names==k) <= length(class_names)) {
        for(j in 1:length(xall_names)){
          Beta[[j]][which(class_names==k),i] <- beta_shell[,k][j] 
        }
        
        if( sum(ck_ind) > 1) {
          bx = t(as.matrix(beta_shell[,k], ncol=1)) %*% t(x_ck) 
          bx = ifelse(bx< -10, -10, ifelse(bx>10, 10, bx))
          
          lk_exist <- lk_exist 
          
        } else if ( sum(ck_ind) == 1 ) {
          bx = t(as.matrix(beta_shell[,k], ncol=1)) %*% x_ck  # beta0 + xi * beta1
          bx = ifelse(bx< -10, -10, ifelse(bx>10, 10, bx))
          lk_exist <- lk_exist 
        }
        
      }
      
      loglik[ck_ind] <- lk_exist
    }
    
    LLmat[,i] <- loglik  
    BIC[i] <- -2*sum(loglik) + 2*length(class_names)*(length(xall_names_num)+1)*log(length(y))
    AIC[i] <- -2*sum(loglik) + 4*length(class_names)*(length(xall_names_num)+1)
    
    
    if( i > burnin ){
      ###--------------------------------------------------------------------###
      #### 5.0 - Predictions on a Training and Test set & Store Results     ####
      ###--------------------------------------------------------------------###
      
      if(!is.null(d_test)){
        name_new_test<-paste0('c', max(as.numeric(substr(class_names,2,10)))+1)
        
        for(p in xall_names_bin){
          c_b_new_test[p] <- rbeta(n = 1, shape1 = 1, shape2 = 1)
        }
        
        for(p in xall_names_num){
          c_n_new_test[1, p] <- rnorm(n = 1, prior_means[p], sqrt(prior_var[p]) )
          c_n_new_test[2, p] <- invgamma::rinvgamma(n = 1,shape = g2[p], rate = b2[p])
        }
        
        beta_new <- mvtnorm::rmvnorm(n = 1, mean = beta_prior_mean,
                                     sigma = diag(beta_prior_var) )          
        
        weights_test <- class_update_test_pdp(n = nt, K = length(class_names) ,
                                              alpha = a,
                                              name_new = name_new_test,
                                              uniq_clabs = colnames(beta_shell),
                                              clabs = c_shell[,1],
                                              x = xt,
                                              x_cat_shell = c_b_shell,
                                              x_num_shell = c_n_shell,
                                              cat_idx = xall_names_bin,   
                                              num_idx = xall_names_num,
                                              cat_new = c_b_new_test,
                                              num_new = c_n_new_test)
        
        c_shell_test <- apply(weights_test, 1, FUN = sample,
                              x=c(colnames(beta_shell), name_new_test),
                              size=1, replace=T)                           
        
        test_pred <- post_pred_draw_test_pdp(n=nt, x=xt, pc=c_shell_test,
                                             beta_shell = beta_shell,
                                             name_new = name_new_test,  
                                             beta_new = beta_new)       
        
        train_pred <- post_pred_draw_train_pdp(n=n, x=x, pc=c_shell[,1],       
                                               beta_shell = beta_shell)          
        
        cluster_inds[['train']][, i-burnin] <- c_shell[,1]
        cluster_inds[['test']][, i-burnin] <- c_shell_test
        
        predictions[['train']][, i-burnin ] <- train_pred
        predictions[['test']][, i-burnin  ] <- test_pred
        
      }else{
        train_pred<-post_pred_draw_train_pdp(n=n, x=x, pc=c_shell[,1],
                                             beta_shell = beta_shell)
        
        cluster_inds[['train']][, i-burnin] <- c_shell[,1]
        
        predictions[['train']][, i-burnin ] <- train_pred
        
      }
    }   
  }     
  
  BIC <- BIC[(burnin+1):iter]
  AIC <- AIC[(burnin+1):iter]
  LLmat <- LLmat[,(burnin+1):iter]
  results <- list(cluster_inds = cluster_inds,
                  predictions = predictions, LLmat = LLmat, AIC=AIC, BIC=BIC)
  cat('Sampling Complete.')
  return(results)
}
